/*
 *  sys_windows.cpp - System dependent routines for Win32
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 *
 *  Ported to Windows by Lauri Pesonen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "sysdeps.h"
#include "main.h"
#include "prefs.h"
#include "user_strings.h"
#include "cpu_emulation.h"
#include "sys.h"
#include "errno.h"
#include "cdenable\ntcd.h"
#include "cdenable\cache.h"
#include "cdenable\eject_nt.h"
#include "cdenable\eject_w9x.h"
#include "cdenable\vxdiface.h"
#include "winioctl.h"
#include "main_windows.h"
#include "sys_windows.h"
#include <mmsystem.h>
#include "cd_defs.h"

#define DEBUG 0
#include "debug.h"

#define MAX_DEVICE_NAME _MAX_PATH
#define DUMMY_HANDLE (HANDLE)-2


// File handles are pointers to these structures
struct file_handle {
	HANDLE h;
	bool is_file;			// Flag: plain file or /dev/something?
	bool is_floppy;		// Flag: floppy device
	bool is_cdrom;		// Flag: CD-ROM device
	bool is_hd;				// Flag: Hard disk.
	bool is_physical;	// Flag: drive instead of partition.
	bool read_only;		// Copy of Sys_open() flag
	off_t start_byte;	// Size of file header (if any)
	cachetype cache;
	char name[MAX_DEVICE_NAME];
	bool is_media_present;
	int drive;
};


static bool is_nt = false;
static bool nocdrom = false;
static bool nofloppyboot = false;

#define FLOPPY_ALIGN_MEMORY_SIZE 512
#define HD_ALIGN_MEMORY_SIZE 512
#define CD_READ_AHEAD_SECTORS 32

static char *sector_buffer = 0;

static bool some_media_removed = false;
static bool some_media_arrived = false;


static void check_os(void)
{
	OSVERSIONINFO osv;

	is_nt = false;

	osv.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if(GetVersionEx( &osv )) {
		switch(osv.dwPlatformId) {
			case VER_PLATFORM_WIN32s:
				ErrorAlert("Cannot run on Win32s");
				QuitEmulator();
				// not reached
				break;
			case VER_PLATFORM_WIN32_NT:
				is_nt = true;
				break;
			default:
				is_nt = false;
		}
	}
}

// A delayed cd handle open mechanism. Makes sure that
// the handle was created in the cpu thread context.

static bool check_cd(file_handle *fh)
{
	if(fh->h) return true;

	if(!is_nt) return(0);

	char dname[100];

	sprintf(dname,"\\\\.\\%c:",(char)*fh->name);
	fh->h = CreateFile (dname, GENERIC_READ,
								 FILE_SHARE_READ|FILE_SHARE_WRITE,
								 NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
								 NULL);
	if( fh->h == INVALID_HANDLE_VALUE ) fh->h = 0;
	return( fh->h != 0 );
}


// Must give cd some time to settle
// Cant give too much however, would be annoying, this is difficult..

static int cd_read_with_retry( file_handle *fh, ULONG LBA, int count, char *buf )
{
	int got_bytes = 0;

	if(!check_cd(fh)) return(0);
	if(is_nt) {
		got_bytes = CdenableSysReadCdBytes( fh->h, LBA, count, buf );
	} else {
		got_bytes = VxdReadCdSectors( fh->drive, LBA, count, buf );
	}
	return(got_bytes);
}

static int cd_read( 
	file_handle *fh, 
	cachetype *cptr,
	ULONG LBA, 
	int count, 
	char *buf
)
{
	ULONG l1, l2, cc;
	int i, c_count, got_bytes = 0, nblocks, s_inx, ss, first_block;
	int ok_bytes = 0;
	char *ptr, *ttptr = 0, *tmpbuf;

	if(fh->h == 0 || count <= 0) return(0);
	if(!check_cd(fh)) return(0);

	ss = 2048;
	l1 = (LBA / ss) * ss;
	l2 = ((LBA+count-1+ss)/ss) * ss;
	cc = l2-l1;
	nblocks = cc / ss;
	first_block = LBA / ss;

	ptr = buf;
	s_inx = LBA-l1;
	c_count = ss - s_inx;
	if(c_count > count) c_count = count;

	for(i=0; i<nblocks; i++) {
		if(cache_get( cptr, first_block+i, sector_buffer )) {
			memcpy( ptr, sector_buffer+s_inx, c_count );
			ok_bytes += c_count;
			ptr += c_count;
			s_inx = 0;
			c_count = ss;
			if(c_count > count-ok_bytes) c_count = count-ok_bytes;
		} else {
			break;
		}
	}

	if(i != nblocks && count != ok_bytes) {
		int bytes_left, blocks_left, alignedleft;

		bytes_left = count - ok_bytes;
		blocks_left = nblocks - i;

		// NEW read ahead code:
		int ahead = CD_READ_AHEAD_SECTORS;
		if(blocks_left < ahead) {
			nblocks += (ahead - blocks_left);
			blocks_left = ahead;
		}

		alignedleft = blocks_left*ss;

		tmpbuf = (char *)VirtualAlloc( 
				NULL, alignedleft, 
				MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
		if(tmpbuf) {
			got_bytes = cd_read_with_retry( fh, (first_block+i)*ss, alignedleft, tmpbuf );
			if(got_bytes != alignedleft) {
				// should never happen
				// Yes it does ...
				if(got_bytes < 0) got_bytes = 0;
				if(c_count > got_bytes) c_count = got_bytes;
				if(c_count > 0) {
					ttptr = tmpbuf;
					memcpy( ptr, ttptr+s_inx, c_count );
					ok_bytes += c_count;
				}
				VirtualFree( tmpbuf, 0, MEM_RELEASE  ); 
				return(ok_bytes); 
			}
			ttptr = tmpbuf;
			for( ; i<nblocks; i++ ) {
				if(c_count > 0) {
					memcpy( ptr, ttptr+s_inx, c_count );
					ok_bytes += c_count;
					ptr += c_count;
				}
				s_inx = 0;
				c_count = ss;
				if(c_count > count-ok_bytes) c_count = count-ok_bytes;
				cache_put( cptr, first_block+i, ttptr, ss );
				ttptr += ss;
			}
			VirtualFree( tmpbuf, 0, MEM_RELEASE  ); 
		}
	}

	return(ok_bytes);
}

size_t nt_floppy_read( file_handle *fh, off_t LBA, size_t count, char *buf )
{
	DWORD bytes_read;
	OVERLAPPED overl;
	char *aux_buffer = 0;
	char *rd_buffer;

	if(!fh->h) return(0);

	if(count == 0) return(0);

	// If not aligned to sector boundary -> must use another buffer
	if( (unsigned long)buf & (FLOPPY_ALIGN_MEMORY_SIZE-1) ) {
		aux_buffer = (char *)VirtualAlloc( 
				NULL, count, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
		if(!aux_buffer) return(0);
		rd_buffer = aux_buffer;
	} else {
		rd_buffer = buf;
	}

	overl.Offset = LBA;
	overl.OffsetHigh = 0;
	overl.hEvent = NULL;

  if (!ReadFile (fh->h, rd_buffer, count, &bytes_read, &overl)) {
		bytes_read = count = 0;
	}
	if(aux_buffer) {
		if(count) memcpy( buf, rd_buffer, count );
		VirtualFree( aux_buffer, 0, MEM_RELEASE  ); 
	}
	return(bytes_read);
}

size_t w9x_floppy_read( file_handle *fh, off_t LBA, size_t count, char *buf )
{
	return(0);
}

size_t w9x_floppy_write( file_handle *fh, ULONG LBA, int count, char *buf )
{
	return(0);
}

size_t floppy_read( file_handle *fh, off_t LBA, size_t count, char *buf )
{
	if(is_nt) {
		return(nt_floppy_read( fh, LBA, count, buf ));
	} else {
		return(w9x_floppy_read( fh, LBA, count, buf ));
	}
}

size_t nt_floppy_write( file_handle *fh, ULONG LBA, int count, char *buf )
{
	DWORD bytes_written;
	OVERLAPPED overl;
	char *aux_buffer = 0;

	if(!fh->h) return(0);
	if(count == 0) return(0);

	overl.Offset = LBA;
	overl.OffsetHigh = 0;
	overl.hEvent = NULL;

	// If not aligned to sector boundary -> must use another buffer
	if( (unsigned long)buf & (FLOPPY_ALIGN_MEMORY_SIZE-1) ) {
		aux_buffer = (char *)VirtualAlloc( 
				NULL, count, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
		if(!aux_buffer) return(0);
		memcpy( aux_buffer, buf, count );
		buf = aux_buffer;
	}
	
  if (WriteFile (fh->h, buf, count, &bytes_written, &overl) &&
			(int)bytes_written == count)
	{
		;
	} else {
		bytes_written = 0;
	}
	if(aux_buffer) VirtualFree( aux_buffer, 0, MEM_RELEASE  ); 
	return(bytes_written);
}

size_t floppy_write( file_handle *fh, off_t LBA, size_t count, char *buf )
{
	if(is_nt) {
		return(nt_floppy_write( fh, LBA, count, buf ));
	} else {
		return(w9x_floppy_write( fh, LBA, count, buf ));
	}
}

/*
 *  Initialization
 */

void SysInit(void)
{
	check_os();

	nocdrom = PrefsFindBool("nocdrom");
	nofloppyboot = PrefsFindBool("nofloppyboot");

	sector_buffer = (char *)VirtualAlloc( NULL, 8192, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );

 	if(is_nt) {
		CdenableSysInstallStart();
	} else {
		if(!VxdInit()) nocdrom = true;
		VxdPatch(1);
	}
}

/*
 *  Deinitialization
 */

void SysExit(void)
{
	if(sector_buffer) {
		VirtualFree( sector_buffer, 0, MEM_RELEASE  ); 
		sector_buffer = 0;
	}
 	if(!is_nt) {
		VxdFinal();
	}
}

static void scan_floppies( char *type )
{
	char rootdir[20], letter;
	int i;

	for( letter = 'A'; letter <= 'B'; letter++ ) {
		i = (int)( letter - 'A' );
		wsprintf( rootdir, "%c:\\", letter );
		if(GetDriveType( rootdir ) == DRIVE_REMOVABLE) {
			PrefsAddString(type, rootdir);
		}
	}
}

/*
 *  This gets called when no "floppy" prefs items are found
 *  It scans for available floppy drives and adds appropriate prefs items
 */

void SysAddFloppyPrefs(void)
{
	scan_floppies("floppy");
}

/*
 *  This gets called when no "disk" prefs items are found
 *  It scans for available HFS volumes and adds appropriate prefs items
 */

// TODO: we shouldn't boot just from any arbitrary volume which happens to
// be the first one. Set up a dialog here, and prompt for a boot volume file,
// and put it first in the list.

void enum_hard_files( char *dir, char *extension )
{
	char mask[_MAX_PATH], path[_MAX_PATH];
	HANDLE fh;
	WIN32_FIND_DATA FindFileData;
	int ok;

	wsprintf( mask, "%s*.%s", dir, extension );
	fh = FindFirstFile( mask, &FindFileData );
	ok = fh != INVALID_HANDLE_VALUE;
	while(ok) {
		ok = FindNextFile( fh, &FindFileData );
		sprintf( path, "%s%s", dir, FindFileData.cFileName );
		PrefsAddString("disk", path);
	}
	if(fh != INVALID_HANDLE_VALUE) FindClose( fh );
}

void enum_hd_partitions(void)
{
	char rootdir[20], letter;
	int i;
	UINT type;

	for( letter = 'C'; letter <= 'Z'; letter++ ) {
		i = (int)( letter - 'A' );
		wsprintf( rootdir, "%c:\\", letter );
		type = GetDriveType( rootdir );
		if(type == DRIVE_FIXED) {
			PrefsAddString("disk", rootdir);
		}
	}
}

void SysAddDiskPrefs(void)
{
	char dir[_MAX_PATH];

	GetCurrentDirectory(sizeof(dir),dir);
	if(strlen(dir) > 0 && dir[strlen(dir)-1] != '\\') strcat(dir,"\\");

	enum_hard_files( dir, "hfv" );
	enum_hard_files( dir, "dsk" );
	enum_hd_partitions();
}


/*
 *  This gets called when no "cdrom" prefs items are found
 *  It scans for available CD-ROM drives and adds appropriate prefs items
 */

void SysAddCDROMPrefs(void)
{
	// Don't scan for drives if nocdrom option given
	if (!nocdrom) {
		char rootdir[20], letter;
		int i;

		for( letter = 'C'; letter <= 'Z'; letter++ ) {
			i = (int)( letter - 'A' );
			wsprintf( rootdir, "%c:\\", letter );
			if(GetDriveType( rootdir ) == DRIVE_CDROM) {
				PrefsAddString("cdrom", rootdir);
			}
		}
	}
}


/*
 *  Add default serial prefs (must be added, even if no ports present)
 */

void SysAddSerialPrefs(void)
{
	PrefsAddString("seriala", "COM1");
	PrefsAddString("serialb", "COM2");
	// This really does not belong here but is the only place
	// I can use w/o modifying code outside of "windows" folder
	PrefsAddBool("nofpu", true);
}

static bool is_floppy_readable( file_handle *fh )
{
	UINT prevmode = SetErrorMode(SEM_NOOPENFILEERRORBOX|SEM_FAILCRITICALERRORS);
	char dummy[0x200];
	bool result = false;

	if( fh && fh->h != INVALID_HANDLE_VALUE ) {
		if(floppy_read( fh, 0, sizeof(dummy), dummy ) == sizeof(dummy)) {
			result = true;
		}
  }
	SetErrorMode(prevmode);
	return( result );
}

static bool is_cdrom_readable( file_handle *fh )
{
	bool result = false;

	if(!check_cd(fh)) return false;

	cache_clear( &fh->cache );
	if(is_nt) {
		DWORD dwBytesReturned;
		result = ( 0 != DeviceIoControl( 
			fh->h,
			IOCTL_STORAGE_CHECK_VERIFY,
			NULL, 0,
			NULL, 0,
			&dwBytesReturned,
			NULL
		) );
	} else {
		// This does not work for CDDA
		size_t bytes = 2048;
		char *buffer = (char *)VirtualAlloc( NULL, bytes, MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE );
		if(buffer) {
			result = (cd_read_with_retry( fh, 0, 2048, buffer ) == 2048);
			VirtualFree( buffer, 0, MEM_RELEASE  ); 
		}
	}
	return(result);
}

// TODO: THESE ARE NOT IMPLEMENTED PROPERLY.
void media_removed(void)
{
	some_media_removed = true;
}

void media_arrived(void)
{
	some_media_arrived = true;
}


/*
 *  Open file/device, create new file handle (returns NULL on error)
 */

// TODO: clean up this function. (too long to be readable)

void *Sys_open(const char *_name, bool read_only)
{
	HANDLE h;
	off_t size;
	file_handle *fh = 0;

	char dnamebuf[64];
	char dname[64];
	char dnamel[64];
	char *name, local_name[_MAX_PATH];

	bool physical = false; // open physical/logical volume

	strcpy( local_name, _name );
	strupr( local_name );

	name = local_name;

	if(strncmp(name,"PHYSICAL ",9) == 0) {
		physical = true;
		name += 9;
		while(*name == ' ') name++;
	}

	// Normalize floppy / cd path.
	// User may give "disk A:" or "disk A:\" or "disk A" or "disk physical A:"

	int name_len = strlen(name);

	if(name_len == 1 && isalpha(*name)) {
		strcat( name, ":\\" );
	}

	if(name_len > 0 && name[name_len-1] == ':') {
		strcat( name, "\\" );
	}

	name_len = strlen(name);

	D(bug("Sys_open(%s, %s)\n", name, read_only ? "read-only" : "read/write"));

	if(name_len > 0 && name[name_len-1] == '\\') {
		UINT type = GetDriveType(name);
		if(type == DRIVE_FIXED) {

			// !!!!!! change this at your own risk.
			read_only = true;

			if(is_nt) {
				*dname = 0;
				if(physical) {
					sprintf(dname,"%c:",(char)toupper(*name));
					if ( QueryDosDevice(dname, dnamebuf, sizeof(dnamebuf)) == 0) {
						return(0);
					} else {
						char *h, *p;
						h = strstr( dnamebuf, "\\Harddisk" );
						p = strstr( dnamebuf, "\\Partition1" );
						if(h == 0 || p == 0) return(0);
						h += 9;
						int inx = atoi(h);
						sprintf( dnamebuf, "\\\\.\\PHYSICALDRIVE%d", inx );
					}
				} else {
					sprintf( dnamebuf, "\\\\.\\%c:", (char)toupper(*name) );
				}

				h = CreateFile( 
					dnamebuf, 
					read_only ? GENERIC_READ : GENERIC_READ|GENERIC_WRITE,
					FILE_SHARE_READ | FILE_SHARE_WRITE,
					0, 
					OPEN_EXISTING, 
					FILE_FLAG_NO_BUFFERING,
					0 
				);
				if( read_only == 0 && h == INVALID_HANDLE_VALUE ) {
					read_only = true;
					h = CreateFile( 
						dnamebuf, 
						GENERIC_READ, 
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						0, 
						OPEN_EXISTING, 
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED,
						0 
					);
					if( h != INVALID_HANDLE_VALUE ) {
						char msg[200];
						wsprintf( 
							msg, 
							"Could not get read/write access to the drive \"%s\", mounting read-only.",
							fh->name
						);
						WarningAlert( msg );
					}
				}
				
				if( read_only == 0 && h != INVALID_HANDLE_VALUE ) {
					DWORD dwBytesReturned = 0;
					if(!DeviceIoControl(h,
						FSCTL_LOCK_VOLUME,
						NULL, 0,
						NULL, 0,
						&dwBytesReturned,
						NULL))
					{
						CloseHandle(h);
						h = INVALID_HANDLE_VALUE;
					}
				}

				if( h != INVALID_HANDLE_VALUE ) {
					fh = new file_handle;
					memset( fh, 0, sizeof(file_handle) );
					fh->h = h;
					fh->is_hd = true;
					fh->is_physical = physical;
					fh->read_only = read_only;
					strcpy( fh->name, name );
					strcpy( fh->name, name );
				}
			} else {
				if(physical) {
					fh = new file_handle;
					memset( fh, 0, sizeof(file_handle) );
					fh->h = DUMMY_HANDLE;
					fh->is_hd = true;
					fh->is_physical = true;

					// If you want r/w access, you MUST grab level 0 lock first.
					fh->read_only = true; // read_only;

					strcpy( fh->name, name );
					fh->drive = toupper(*name) - 'A';
				}
			}
		} else if(GetDriveType(name) == DRIVE_REMOVABLE) {
			if(is_nt) {
				int rc;
				sprintf(dname,"Basilisk2_floppy_%c",(char)toupper(*name));

				// TODO: fails on systems where \Device\Floppy0 is B:
				sprintf(dnamel,"\\Device\\Floppy%d",(int)toupper(*name) - 'A');

				if ( QueryDosDevice(dname, dnamebuf, sizeof(dnamebuf)) == 0) {
					rc = GetLastError();
					if ( rc == ERROR_FILE_NOT_FOUND ) {
						if ( DefineDosDevice(DDD_RAW_TARGET_PATH, dname, dnamel) == 0 ) {
							return( 0);
						}
					} else {
						return(0);
					}
				}
				sprintf(dnamebuf,"\\\\.\\%s", dname);

				UINT prevmode = SetErrorMode(SEM_NOOPENFILEERRORBOX|SEM_FAILCRITICALERRORS);

				sprintf(dname,"\\\\.\\%c:",(char)*name);
				h = CreateFile (dname, GENERIC_READ|GENERIC_WRITE,
											 FILE_SHARE_READ,
											 NULL, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS,
											 0);
				if( h != INVALID_HANDLE_VALUE ) {
					fh = new file_handle;
					memset( fh, 0, sizeof(file_handle) );
					fh->h = h;
					fh->is_floppy = true;
					fh->read_only = read_only;
					strcpy( fh->name, name );
					// memset( &fh->cache, 0, sizeof(cachetype) );
					// cache_init( &fh->cache );
					if (!nofloppyboot) {
						fh->is_media_present = is_floppy_readable(fh);
					}
  			}
				SetErrorMode(prevmode);
			} else {
				fh = new file_handle;
				memset( fh, 0, sizeof(file_handle) );
				fh->h = DUMMY_HANDLE;
				fh->is_floppy = true;
				fh->read_only = read_only;
				strcpy( fh->name, name );
				fh->drive = toupper(*name) - 'A';
				if (!nofloppyboot) {
					fh->is_media_present = is_floppy_readable(fh);
				}
			}
		} else if(GetDriveType(name) == DRIVE_CDROM) {
			read_only = true;
			if(is_nt) {
				sprintf(dname,"\\\\.\\%c:",(char)*name);
				h = CreateFile (dname, GENERIC_READ,
											 FILE_SHARE_READ|FILE_SHARE_WRITE,
											 NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
											 NULL);
				if( h != INVALID_HANDLE_VALUE ) {
					fh = new file_handle;
					memset( fh, 0, sizeof(file_handle) );
					fh->h = h;
					fh->is_cdrom = true;
					fh->read_only = read_only;
					fh->drive = toupper(*name) - 'A';
					strcpy( fh->name, name );
					memset( &fh->cache, 0, sizeof(cachetype) );
					cache_init( &fh->cache );
					if (!nocdrom) {
						fh->is_media_present = is_cdrom_readable(fh);
					}
					// Need to close the handle and reopen it later.
					// We are in the GUI thread context now.
					CloseHandle(fh->h);
					fh->h = 0;
				}
			} else {
				fh = new file_handle;
				memset( fh, 0, sizeof(file_handle) );
				fh->h = DUMMY_HANDLE;
				fh->is_cdrom = true;
				fh->read_only = read_only;
				fh->drive = toupper(*name) - 'A';
				strcpy( fh->name, name );
				memset( &fh->cache, 0, sizeof(cachetype) );
				cache_init( &fh->cache );
				if (!nocdrom) {
					fh->is_media_present = is_cdrom_readable(fh);
				}
			}
		}
	} else {
		// hard file.
		// Check if write access is allowed, set read-only flag if not
		// Open file/device (for exclusive access, to protect data integrity)
		h = CreateFile( 
			name,
			GENERIC_READ | GENERIC_WRITE,
			0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL
		);

		if(h == INVALID_HANDLE_VALUE) {
			read_only = true;
			h = CreateFile( 
				name,
				GENERIC_READ,
				0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL
			);
		}

		if(h != INVALID_HANDLE_VALUE) {
			fh = new file_handle;
			memset( fh, 0, sizeof(file_handle) );
			fh->h = h;
			fh->is_file = true;
			fh->read_only = read_only;
			strcpy( fh->name, name );
			if (fh->is_file) {
				// Detect header length: if the file size is not a multiple of 512,
				// the remainder at the beginning is considered the header and skipped
				size = (long)GetFileSize(h,NULL);
				if (size == 838484)	// 800K DiskCopy image
					fh->start_byte = 84;
				else
					fh->start_byte = size % 512;
			}
		}
	}

	return fh;
}


/*
 *  Close file/device, delete file handle
 */

void Sys_close(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (fh) {
		if( fh->is_cdrom  ) {
			cache_final( &fh->cache );
		}
		if( !fh->read_only && fh->is_hd && fh->h ) {
			DWORD dwBytesReturned = 0;
			DeviceIoControl(
				fh->h,
				FSCTL_UNLOCK_VOLUME,
				NULL, 0,
				NULL, 0,
				&dwBytesReturned,
				NULL
			);
		}
		if( fh->h != 0 && fh->h != DUMMY_HANDLE) {
			CloseHandle( fh->h );
			fh->h = 0;
		}
		delete fh;
	}
}

size_t nt_hd_read( file_handle *fh, loff_t offset, int count, char *buf )
{
	DWORD bytes_read;
	OVERLAPPED overl;
	char *aux_buffer = 0;
	char *rd_buffer;

	if(!fh->h) return(0);

	if(count == 0) return(0);

	// If not aligned to sector boundary -> must use another buffer
	if( (unsigned long)buf & (HD_ALIGN_MEMORY_SIZE-1) ) {
		aux_buffer = (char *)VirtualAlloc( 
				NULL, count, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
		if(!aux_buffer) return(0);
		rd_buffer = aux_buffer;
	} else {
		rd_buffer = buf;
	}

	overl.Offset = (DWORD)offset;
	overl.OffsetHigh = (DWORD)(offset >> 32);
	overl.hEvent = NULL;

  if (!ReadFile (fh->h, rd_buffer, count, &bytes_read, &overl)) {
		bytes_read = count = 0;
	}
	if(aux_buffer) {
		if(bytes_read) memcpy( buf, rd_buffer, bytes_read );
		VirtualFree( aux_buffer, 0, MEM_RELEASE  ); 
	}
	return(bytes_read);
}

size_t w9x_hd_read( file_handle *fh, loff_t offset, int count, char *buf )
{
	return( VxdReadHdSectors( fh->drive, (DWORD)offset, count, buf ) );
}

size_t hd_read( file_handle *fh, loff_t offset, int count, char *buf )
{
	if(is_nt) {
		return( nt_hd_read( fh, offset, count, buf ) );
	} else {
		return( w9x_hd_read( fh, offset, count, buf ) );
	}
}

int nt_hd_write( file_handle *fh, loff_t offset, int count, char *buf )
{
	DWORD bytes_written;
	OVERLAPPED overl;
	char *aux_buffer = 0;
	char *rd_buffer;

	if(!fh->h) return(0);

	if(count == 0) return(0);

	// If not aligned to sector boundary -> must use another buffer
	if( (unsigned long)buf & (HD_ALIGN_MEMORY_SIZE-1) ) {
		aux_buffer = (char *)VirtualAlloc( 
				NULL, count, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
		if(!aux_buffer) return(0);
		rd_buffer = aux_buffer;
		memcpy( rd_buffer, buf, count );
	} else {
		rd_buffer = buf;
	}

	overl.Offset = (DWORD)offset;
	overl.OffsetHigh = (DWORD)(offset >> 32);
	overl.hEvent = NULL;

  if (!WriteFile (fh->h, rd_buffer, count, &bytes_written, &overl)) {
		if(GetLastError() == ERROR_WRITE_PROTECT) {
			// AfxMessageBox( "The media is write protected." );
		}
		bytes_written = count = 0;
	}

	if(aux_buffer) {
		VirtualFree( aux_buffer, 0, MEM_RELEASE  ); 
	}
	return(bytes_written);
}

int w9x_hd_write( file_handle *fh, loff_t offset, int count, char *buf )
{
	// return( VxdWriteHdSectors( fh->drive, LBA, count, buf ) );
	return( 0 );
}

int hd_write( file_handle *fh, loff_t offset, int count, char *buf )
{
	if(is_nt) {
		return(nt_hd_write( fh, offset, count, buf ));
	} else {
		return(w9x_hd_write( fh, offset, count, buf ));
	}
}

/*
 *  Read "length" bytes from file/device, starting at "offset", to "buffer",
 *  returns number of bytes read (or 0)
 */

size_t Sys_read(void *arg, void *buffer, loff_t offset, size_t length)
{
	file_handle *fh = (file_handle *)arg;
	DWORD bytes_read = 0;
	LONG low, high;

	if (!fh) return 0;

	if(fh->is_file) {
		offset += fh->start_byte;
		low = (LONG)offset;
		high = (LONG)(offset >> 32);
		if( SetFilePointer( fh->h, low, &high, FILE_BEGIN ) != 0xFFFFFFFF ) {
			if(!ReadFile( fh->h, buffer, length, &bytes_read, NULL)) {
				bytes_read = 0;
			}
		}
	} else if(fh->is_floppy || fh->is_cdrom || fh->is_hd) {
		if(fh->is_cdrom && !check_cd(fh)) return 0;
		size_t bytes_left, try_bytes, got_bytes;
		char *b = (char *)buffer;
		bytes_left = length;
		while(bytes_left) {
			try_bytes = min( bytes_left, 32768 );
			if(fh->is_cdrom) {
				got_bytes = cd_read( fh, &fh->cache, (DWORD)offset, try_bytes, b );
				if(got_bytes != try_bytes) {
					if (!nocdrom) {
						fh->is_media_present = is_cdrom_readable(fh);
					}
				}
			} else if(fh->is_floppy) {
				got_bytes = floppy_read( fh, (DWORD)offset, try_bytes, (char*)buffer );
				if(got_bytes != try_bytes) {
					fh->is_media_present = is_floppy_readable(fh);
				}
			} else if(fh->is_hd) {
				got_bytes = hd_read( fh, offset, try_bytes, (char*)b );
			}
			b += got_bytes;
			offset += got_bytes;
			bytes_read += got_bytes;
			bytes_left -= got_bytes;
			if(got_bytes != try_bytes) {
				bytes_left = 0;
			}
		}
	}

	return(bytes_read);
}


/*
 *  Write "length" bytes from "buffer" to file/device, starting at "offset",
 *  returns number of bytes written (or 0)
 */

size_t Sys_write(void *arg, void *buffer, loff_t offset, size_t length)
{
	file_handle *fh = (file_handle *)arg;
	DWORD bytes_written = 0;
	LONG low, high;

	if(!fh) return 0;

	if(fh->is_file) {
		offset += fh->start_byte;
		low = (LONG)offset;
		high = (LONG)(offset >> 32);
		if( SetFilePointer( fh->h, low, &high, FILE_BEGIN ) != 0xFFFFFFFF ) {
			if(!WriteFile( fh->h, buffer, length, &bytes_written, NULL)) {
				bytes_written = 0;
			}
		}
	} else if(fh->is_hd) {
		if(fh->read_only) {
			D(bug("Attempt to write to read-only HD partition rejected.\r\n"));
		} else {
			bytes_written = hd_write( fh, offset, length, (char *)buffer );
		}
	} else if(fh->is_floppy) {
		bytes_written = floppy_write( fh, (DWORD)offset, length, (char *)buffer );
		if(bytes_written != length) {
			fh->is_media_present = is_floppy_readable(fh);
		}
	}
	return(bytes_written);
}

/*
 *  Return size of file/device (minus header)
 */

loff_t SysGetFileSize(void *arg)
{
	off_t length = 0;
	LONGLONG real_len = 0;

	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return true;

	if(fh->h) {
		if (fh->is_file) {
			real_len = (LONGLONG)GetFileSize(fh->h,NULL);
		} else if (fh->is_floppy) {
			real_len = (LONGLONG)80 * 2 * 18 * 512;
		} else if (fh->is_cdrom) {
			if(!check_cd(fh)) return 0;
			// test
			real_len = (LONGLONG)0x28A00000;
		} else if (fh->is_hd) {
			// Note: if pure mac disk, cluster info is not available
			if(win_os_old) {
				DWORD SectorsPerCluster;
				DWORD BytesPerSector;
				DWORD NumberOfFreeClusters;
				DWORD TotalNumberOfClusters;
				if(GetDiskFreeSpace( fh->name,
					&SectorsPerCluster,
					&BytesPerSector,
					&NumberOfFreeClusters,
					&TotalNumberOfClusters ))
				{
					real_len = (LONGLONG)TotalNumberOfClusters*SectorsPerCluster*BytesPerSector;
				}
			} else if(fh->is_physical) {
				DISK_GEOMETRY g;
				DWORD dwBytesReturned;
				if(DeviceIoControl( 
					fh->h,
					IOCTL_DISK_GET_DRIVE_GEOMETRY,
					NULL, 0,
					&g, sizeof(g),
					&dwBytesReturned,
					NULL
				) && dwBytesReturned == sizeof(g) ) 
				{
					real_len = ( (LONGLONG)g.Cylinders.QuadPart *
											 (LONGLONG)g.TracksPerCylinder *
											 (LONGLONG)g.SectorsPerTrack *
											 (LONGLONG)g.BytesPerSector );
				}
			} else {
				// TODO: use loadlibrary() ... getprocaddress().
				/*
				ULARGE_INTEGER FreeBytesAvailableToCaller; // Note: quotas.
				ULARGE_INTEGER TotalNumberOfBytes;
				ULARGE_INTEGER TotalNumberOfFreeBytes;
				if(GetDiskFreeSpaceEx( fh->name,
					&FreeBytesAvailableToCaller,
					&TotalNumberOfBytes,
					&TotalNumberOfFreeBytes))
				{
					real_len = (LONGLONG)TotalNumberOfBytes.QuadPart;
				}
				*/
				DWORD SectorsPerCluster;
				DWORD BytesPerSector;
				DWORD NumberOfFreeClusters;
				DWORD TotalNumberOfClusters;
				if(GetDiskFreeSpace( fh->name,
					&SectorsPerCluster,
					&BytesPerSector,
					&NumberOfFreeClusters,
					&TotalNumberOfClusters ))
				{
					real_len = (LONGLONG)TotalNumberOfClusters*SectorsPerCluster*BytesPerSector;
				}
			}
		}
	}

#if DEBUG
if(sizeof(off_t) != 4) D(bug("off_t size changed, update code\r\n"));
#endif

	/*
	if(real_len > (LONGLONG)0x7FFFFFFF) {
		// what else to do
		length = (off_t)0x7FFFFFFF;
	} else {
		length = (off_t)real_len;
	}
	*/

	return real_len;
}

/*
 *  Eject volume (if applicable)
 */

void SysEject(void *arg)
{
	DWORD dwBytesReturned;
	file_handle *fh = (file_handle *)arg;

	if (!fh) return;

	if (fh->is_floppy) {
		char msg[300];
		if(fh->is_media_present) {
			fh->is_media_present = false;
			if(is_nt) {
				DeviceIoControl( 
					fh->h,
					IOCTL_STORAGE_EJECT_MEDIA,
					NULL, 0,
					NULL, 0,
					&dwBytesReturned,
					NULL
				);
			} else {
				EjectMedia_w9x( fh->drive );
			}
			wsprintf( 
				msg, 
				"The floppy was ejected from drive \"%c:\".\r\n"
				"Press Control-Shift-F11 to reload the floppy disk.", 
				// "Hold both left and right control keys down for %d seconds to reload the floppy.", 
				(char)toupper(*fh->name)
				// , get_floppy_detection_seconds()
			);
			WarningAlert( msg );
		}
	} else if (fh->is_cdrom) {
		if(!check_cd(fh)) return;
		fh->is_media_present = false;
		if(is_nt) {

			// Commented out because there was some problems, but can't remember
			// exactly ... need to find out

			// EjectVolume(toupper(*fh->name),false);

			// Preventing is cumulative, try to make sure it's indeed released now
			for(int i=0; i<10; i++) PreventRemovalOfVolume(fh->h,false);

			if (!nocdrom) {
				DeviceIoControl( 
					fh->h,
					IOCTL_STORAGE_EJECT_MEDIA,
					NULL, 0,
					NULL, 0,
					&dwBytesReturned,
					NULL
				);
			}
		} else {
			EjectMedia_w9x( fh->drive + 1 );
		}
		cache_clear( &fh->cache );
	}
}

/*
 *  Format volume (if applicable)
 */

bool SysFormat(void *arg)
{
	file_handle *fh = (file_handle *)arg;

	if (!fh) return false;

	if(fh->read_only) return false;

	return true;
}


/*
 *  Check if file/device is read-only (this includes the read-only flag on Sys_open())
 */

bool SysIsReadOnly(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return true;

	return fh->read_only;
}


/*
 *  Check if the given file handle refers to a fixed or a removable disk
 */

bool SysIsFixedDisk(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return true;

	if (fh->is_file || fh->is_hd)
		return true;
	else if (fh->is_floppy || fh->is_cdrom)
		return false;
	else
		return true;
}


/*
 *  Check if a disk is inserted in the drive (always true for files)
 */

bool SysIsDiskInserted(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	bool result = false;

	if (!fh) return false;

	if (fh->is_file || fh->is_hd) {
		result = true;
	} else if (fh->is_floppy) {
		if(!fh->is_media_present) {
			if (!nofloppyboot) {
				fh->is_media_present = is_floppy_readable(fh);
			} else {
				// if nofloppyboot, only after MacOS started
				if (ReadMacInt32(0xcfc) == 'WLSC') {
					fh->is_media_present = is_floppy_readable(fh);
				}
			}
		}
		result = fh->is_media_present;
	} else if (fh->is_cdrom) {
		if(!check_cd(fh)) return false;
		if(is_nt) {
			fh->is_media_present = is_cdrom_readable(fh);
		} else {
			if(some_media_arrived || some_media_removed) {
				fh->is_media_present = is_cdrom_readable(fh);
				some_media_arrived = false;
				some_media_removed = false;
			}
		}
		result = fh->is_media_present;
		D(bug("is cd inserted = %d\r\n",(int)result));
	}
	return(result);
}


/*
 *  Prevent medium removal (if applicable)
 */

void SysPreventRemoval(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return;

	if (fh->is_cdrom) {
		if(!check_cd(fh)) return;
		if(is_nt) {
			PreventRemovalOfVolume(fh->h,true);
		}
	}
}


/*
 *  Allow medium removal (if applicable)
 */

void SysAllowRemoval(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return;

	if (fh->is_cdrom) {
		if(!check_cd(fh)) return;
		if(is_nt) {
			for( int i=0; i<10; i++) PreventRemovalOfVolume(fh->h,false);
		}
	}
}


// TOC and CDDA for NT only


/*
 *  Read CD-ROM TOC (binary MSF format, 804 bytes max.)
 */

bool SysCDReadTOC(void *arg, uint8 *toc)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return false;
	if(!is_nt) return false;

	DWORD dwBytesReturned = 0;

	// to ease viewing in db
	PCDROM_TOC ptoc = (PCDROM_TOC)toc;

	if(!check_cd(fh)) return false;

	if(DeviceIoControl(fh->h,
		IOCTL_CDROM_READ_TOC,
		NULL, 0,
		toc, sizeof(CDROM_TOC),
		&dwBytesReturned,
		NULL))
	{
		D(bug("SysCDReadTOC success\r\n"));
		return true;
	}
	D(bug("SysCDReadTOC failed\r\n"));
	return false;
}


/*
 *  Read CD-ROM position data (Sub-Q Channel, 16 bytes, see SCSI standard)
 */

bool SysCDGetPosition(void *arg, uint8 *pos)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return false;
	if(!is_nt) return false;
	if(!check_cd(fh)) return false;

	DWORD dwBytesReturned = 0;
	SUB_Q_CHANNEL_DATA q_data;

	if(DeviceIoControl(fh->h,
		IOCTL_CDROM_READ_Q_CHANNEL,
		NULL, 0,
		&q_data, sizeof(SUB_Q_CHANNEL_DATA),
		&dwBytesReturned,
		NULL))
	{
		memcpy( pos, &q_data.CurrentPosition, sizeof(SUB_Q_CURRENT_POSITION) );
		D(bug("SysCDGetPosition success\r\n"));
		return true;
	} else {
		D(bug("SysCDGetPosition failed\r\n"));
		return false;
	}
}


/*
 *  Play CD audio
 */

bool SysCDPlay(void *arg, uint8 start_m, uint8 start_s, uint8 start_f, uint8 end_m, uint8 end_s, uint8 end_f)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return false;
	if(!is_nt) return false;
	if(!check_cd(fh)) return false;

	DWORD dwBytesReturned = 0;
	CDROM_PLAY_AUDIO_MSF msf;

	D(bug("SysCDPlay %d,%d,%d, %d,%d,%d\r\n",(int)start_m,(int)start_s,(int)start_f,(int)end_m,(int)end_s,(int)end_f));

  msf.StartingM = start_m;
  msf.StartingS = start_s;
  msf.StartingF = start_f;
  msf.EndingM = end_m;
  msf.EndingS = end_s;
  msf.EndingF = end_f;

	if(DeviceIoControl(fh->h,
		IOCTL_CDROM_PLAY_AUDIO_MSF,
		&msf, sizeof(CDROM_PLAY_AUDIO_MSF),
		NULL, 0,
		&dwBytesReturned,
		NULL))
	{
		D(bug("SysCDPlay success\r\n"));
		return true;
	} else {
		D(bug("SysCDPlay failed\r\n"));
		return false;
	}
}


/*
 *  Pause CD audio
 */

bool SysCDPause(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return false;
	if(!is_nt) return false;
	if(!check_cd(fh)) return false;

	DWORD dwBytesReturned = 0;
	if(DeviceIoControl(fh->h,IOCTL_CDROM_PAUSE_AUDIO, 0, 0, 0, 0, &dwBytesReturned, 0)) {
		D(bug("SysCDPause success\r\n"));
		return true;
	} else {
		D(bug("SysCDPause failed\r\n"));
		return false;
	}
}


/*
 *  Resume paused CD audio
 */

bool SysCDResume(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return false;
	if(!is_nt) return false;
	if(!check_cd(fh)) return false;

	DWORD dwBytesReturned = 0;
	if(DeviceIoControl(fh->h,IOCTL_CDROM_RESUME_AUDIO, 0, 0, 0, 0, &dwBytesReturned, 0)) {
		D(bug("SysCDResume success\r\n"));
		return true;
	} else {
		D(bug("SysCDResume failed\r\n"));
		return false;
	}
}


/*
 *  Stop CD audio
 */

bool SysCDStop(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return false;
	if(!is_nt) return false;
	if(!check_cd(fh)) return false;

	DWORD dwBytesReturned = 0;
	if(DeviceIoControl(fh->h,IOCTL_CDROM_STOP_AUDIO, 0, 0, 0, 0, &dwBytesReturned, 0)) {
		D(bug("SysCDStop success\r\n"));
		return true;
	} else {
		D(bug("SysCDStop failed\r\n"));
		return false;
	}
}


/*
 *  Perform CD audio fast-forward/fast-reverse operation starting from specified address
 */

bool SysCDScan(void *arg, uint8 start_m, uint8 start_s, uint8 start_f, bool reverse)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return false;
	if(!is_nt) return false;
	if(!check_cd(fh)) return false;

	DWORD dwBytesReturned = 0;
	CDROM_SEEK_AUDIO_MSF msf;

	// reverse?

	D(bug("SysCDScan %d,%d,%d,%d\r\n",(int)start_m, (int)start_s, (int)start_f, (int)reverse));

  msf.M = start_m;
  msf.S = start_s;
  msf.F = start_f;

	if(DeviceIoControl(fh->h,
		IOCTL_CDROM_SEEK_AUDIO_MSF,
		&msf, sizeof(CDROM_SEEK_AUDIO_MSF),
		NULL, 0,
		&dwBytesReturned,
		NULL))
	{
		D(bug("SysCDScan success\r\n"));
		return true;
	} else {
		D(bug("SysCDScan failed\r\n"));
		return false;
	}
}


/*
 *  Set CD audio volume (0..255 each channel)
 */

void SysCDSetVolume(void *arg, uint8 left, uint8 right)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return;
	if(!is_nt) return;
	if(!check_cd(fh)) return;

	DWORD dwBytesReturned = 0;
	VOLUME_CONTROL vc;

	vc.PortVolume[0] = left;
	vc.PortVolume[1] = right;
	vc.PortVolume[2] = left;
	vc.PortVolume[3] = right;

	D(bug("SysCDSetVolume %d,%d\r\n",(int)left,(int)right));

	DeviceIoControl(fh->h,
		IOCTL_CDROM_SET_VOLUME,
		&vc, sizeof(VOLUME_CONTROL),
		NULL, 0,
		&dwBytesReturned,
		NULL);
}


/*
 *  Get CD audio volume (0..255 each channel)
 */

void SysCDGetVolume(void *arg, uint8 &left, uint8 &right)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return;
	if(!is_nt) return;
	if(!check_cd(fh)) return;

	DWORD dwBytesReturned = 0;
	VOLUME_CONTROL vc;

	if(DeviceIoControl(fh->h,
		IOCTL_CDROM_GET_VOLUME,
		NULL, 0,
		&vc, sizeof(VOLUME_CONTROL),
		&dwBytesReturned,
		NULL))
	{
		left = vc.PortVolume[0];
		right = vc.PortVolume[1];
		D(bug("SysCDGetVolume success %d,%d\r\n",(int)left,(int)right));
	} else {
		left = 0;
		right = 0;
		D(bug("SysCDGetVolume failed\r\n"));
	}
}
