/*
 *  video_windows.cpp - Video/graphics emulation for Win32
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 *
 *  Ported to Windows by Lauri Pesonen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "sysdeps.h"
#include "main.h"
#include "adb.h"
#include "prefs.h"
#include "user_strings.h"
#include "cpu_emulation.h"
#include "main_windows.h"
#include "video_windows.h"
#include "desktop_windows.h"
#include "video.h"
#include <winnt.h>
#include <ddraw.h>
#include <d3d.h>
#include <process.h>

#define DEBUG 0
#include "debug.h"


/*
  If you undefine USE_FAULTS, apps like SpeedoMeter may give somewhat
  higher scores. Don't let this fool you, the overall performance is
  much better with USE_FAULTS defined, especially with high color modes.

	This method can be improved further by noting that Mac Finder updates the
	menu bar (about 20 top lines) often. Therefore, if we move mouse at the 
	bottom of the screen, sometimes almost the whole screen must be updated.
	Should be split into two areas.
	Not sure if I will bother, the screen is not the bottleneck any more.
*/

#ifndef _DEBUG
#define USE_FAULTS
#endif

static char *window_class = "Basilisk II MainWndClass";

// For testing purposes only, this one flickers badly.
// #define SHOW_REAL_FRAME_RATE

// Need this as a straight integer
static int screen_bit_depth;

int is_windowed = false;

int is_ontop = 0;

// Usually (but not always) it's best to use palette index
// and not explicit rgb values on 8bit devices
static bool bPalette = true;

// Used to determine do we need to release frame buffer lock.
static bool dx_blits_allowed = true;

// If you have NT 4.0 with DirectX 5.0 hacked in, you may need this.
// It fixes the black/white palette problem by using GDI palette
// instead of D3D palette.
static bool nt_dx5_palette_hack = false;

// How many msecs to sleep between screen refreshes
#define SLEEP_TICKS 30

// Platform specific: size of the virtual memory page.
#define VM_PAGE_BYTES       4096
#define VM_PAGE_BITS        12
#define VM_PAGE_ALIGN_MASK  0xFFFFF000
#define VM_PAGE_MODULO_BITS 0xFFF


// Mac specific: screen size
#define SCREEN_BYTES    scanlines[VideoMonitor.y]
#define SCREEN_MAX_X    VideoMonitor.x
#define SCREEN_MAX_Y    VideoMonitor.y
#define SCANLINE_BYTES  VideoMonitor.bytes_per_row

// How many virtual memory pages per Mac screen?.
// This just needs to be large enough. Should really be calculated
// using display attributes, fix it if you want
#define MAX_PAGES_PER_SCREEN 10000

// Current palette in "MacOS values".
uint8 current_mac_palette[256*3];

int display_type = DISPLAY_WINDOW;

typedef struct {
  int top, bottom;  // Mapping between this virtual page and Mac scan lines
} ScreenPageInfo;

typedef struct {
  DWORD memBase,    // Real start address of this Mac screen page
        memStart,   // Start address aligned to nearest page boundary
        memEnd,     // Address of the first byte not belonging to the screen pages
        memLength;  // Length of the memory addressed by the screen pages
  int dirty,        // Anything written to this buffer?
      top,          // Top scanline of dirty area
      bottom;       // First "clean" scanline
  DWORD unprotect_start; // speedup.
  DWORD unprotect_bytes; // speedup
  ScreenPageInfo pages[MAX_PAGES_PER_SCREEN];
} ScreenInfo;


// Global variables
static bool screen_inited = 0;

static int32 frame_skip = 1;
uint8 *the_buffer = 0; 

static HBITMAP hScreenBitmap = 0;
static HDC hBitMapDC = 0;
static HDC hdcBits = 0;
static HPALETTE hPalMain = 0;
static HPALETTE hPalBits = 0;
static LPBITMAPINFO pbmi = 0;
static LPLOGPALETTE lppal = 0;

static CRITICAL_SECTION draw_csection;
static CRITICAL_SECTION fb_csection;

#define MAX_SCAN_LINES 5000
int scanlines[MAX_SCAN_LINES];
int dx_scanlines[MAX_SCAN_LINES];


// Prototypes
unsigned int WINAPI redraw_thread(LPVOID param);
unsigned int WINAPI redraw_thread_d3d(LPVOID param);
unsigned int WINAPI redraw_thread_fb(LPVOID param);

static uint8 fb_current_command = FB_NONE;
HANDLE fb_signal = 0;
HANDLE fb_reply = 0;

static inline void grab_draw_mutex(void)
{
	EnterCriticalSection( &draw_csection );
}

static inline void release_draw_mutex(void)
{
	LeaveCriticalSection( &draw_csection );
}

#ifdef USE_FAULTS

static ScreenInfo mainBuffer;

// We don't really need this.
static LPTOP_LEVEL_EXCEPTION_FILTER nextFaultFunc = 0;

// This function is called when any exception would remain
// unhandled. It is a write to the Mac page if:
//    - it is a access violation
//    - it is within the range of the Mac screen buffers.
//    - if both of these are true, we know that it's a write
//      (since we didn't deny read).
// Update the dirty area (top,bottom) of the buffer,
// make the virtual page r/w again to stop faulting,
// and continue execution. The write will be attempted again
// and it will succeed now.

LONG WINAPI Screen_fault_proc( struct _EXCEPTION_POINTERS *ExceptionInfo )
{
  DWORD a;
  DWORD OldProtect;
  int page, lines, top, bottom;

  if(ExceptionInfo->ExceptionRecord->ExceptionCode == EXCEPTION_ACCESS_VIOLATION) {
    a = (DWORD)ExceptionInfo->ExceptionRecord->ExceptionInformation[1];
    if(a >= mainBuffer.memStart && a < mainBuffer.memEnd) {
			grab_draw_mutex();
      page = (a - mainBuffer.memStart) >> VM_PAGE_BITS;
      top = mainBuffer.pages[page].top;
      bottom = mainBuffer.pages[page].bottom;

      lines = bottom-top;
      if(lines == 0) lines = 1;

      a = mainBuffer.memStart + scanlines[top];

			// D(bug("top=%d, bottom=%d, dirty=%d\n",top,bottom,mainBuffer.dirty));

      if(mainBuffer.dirty) {
				//D(bug("Was dirty\n"));
        mainBuffer.top    = min( mainBuffer.top, top );
        mainBuffer.bottom = max( mainBuffer.bottom, bottom );
        DWORD b = mainBuffer.unprotect_start+mainBuffer.unprotect_bytes;
        mainBuffer.unprotect_start = min(a,mainBuffer.unprotect_start);
        mainBuffer.unprotect_bytes = max(b,a+scanlines[lines]) - mainBuffer.unprotect_start;
        VirtualProtect( (LPVOID)mainBuffer.unprotect_start, mainBuffer.unprotect_bytes, PAGE_READWRITE, &OldProtect );
      } else {
				//D(bug("Was clean\n"));
        mainBuffer.top    = top;
        mainBuffer.bottom = bottom;
        mainBuffer.dirty  = 1;
        mainBuffer.unprotect_start = a;
        mainBuffer.unprotect_bytes = scanlines[lines];
        VirtualProtect( (LPVOID)a, mainBuffer.unprotect_bytes, PAGE_READWRITE, &OldProtect );
      }
			release_draw_mutex();
      return( EXCEPTION_CONTINUE_EXECUTION );

		// I'm just testing here.
		/*
    } else if(a >= (DWORD)ROMBaseHost && a < (DWORD)ROMBaseHost + ROMSize) {
			LPBYTE eip = (LPBYTE)ExceptionInfo->ContextRecord->Eip;
			// _asm int 3
			if(*eip == 0x88 && eip[1] == 0x18) {
				ExceptionInfo->ContextRecord->Eip += 2;
	      return( EXCEPTION_CONTINUE_EXECUTION );
			} else if(*eip == 0x88) {
				ExceptionInfo->ContextRecord->Eip += 3;
	      return( EXCEPTION_CONTINUE_EXECUTION );
			}
		*/
    }
  }
  return( EXCEPTION_CONTINUE_SEARCH );
}

void Screen_Setup_fault_handler(void)
{
  int i, y1, y2;
  DWORD a;

  mainBuffer.memBase  = (DWORD)the_buffer;
  mainBuffer.memStart = mainBuffer.memBase & VM_PAGE_ALIGN_MASK;
  mainBuffer.memEnd   = mainBuffer.memBase + SCREEN_BYTES;
  if(mainBuffer.memEnd & VM_PAGE_MODULO_BITS) mainBuffer.memEnd = (mainBuffer.memEnd | VM_PAGE_MODULO_BITS) + 1;
  mainBuffer.memLength = mainBuffer.memEnd - mainBuffer.memStart;
  mainBuffer.dirty = 1;
  mainBuffer.top = 0;
  mainBuffer.bottom = SCREEN_MAX_Y;

  mainBuffer.unprotect_start = mainBuffer.memStart;
  mainBuffer.unprotect_bytes = mainBuffer.memLength;

  a = 0;
  for(i=0; i<(unsigned int)MAX_PAGES_PER_SCREEN; i++) {
		y1 = a / (int)SCANLINE_BYTES;
		y2 = (a+VM_PAGE_BYTES) / (int)SCANLINE_BYTES;
		y1 = y1 - 16;
		y2 = y2 + 64;
		if(y1 < 0) y1 = 0;
		if(y2 < 0) y2 = 0;
		if(y1 > (int)VideoMonitor.y) y1 = (int)VideoMonitor.y;
		if(y2 > (int)VideoMonitor.y) y2 = (int)VideoMonitor.y;
    mainBuffer.pages[i].top = y1;
    mainBuffer.pages[i].bottom = y2;
    a += VM_PAGE_BYTES;
  }

  nextFaultFunc = SetUnhandledExceptionFilter( (LPTOP_LEVEL_EXCEPTION_FILTER)Screen_fault_proc );
}
#endif //USE_FAULTS

/*
 *  Initialization
 */

// Set VideoMonitor according to video mode
void set_video_monitor(int width, int height, int depth_win, int depth_mac, int bytes_per_row)
{
	screen_bit_depth = depth_win;

  switch (depth_mac) {
    case 8:
      MacFrameLayout = FLAYOUT_DIRECT;
      VideoMonitor.mode = VMODE_8BIT;
      break;
    case 15:
      MacFrameLayout = FLAYOUT_HOST_555;
      VideoMonitor.mode = VMODE_16BIT;
      break;
    case 16:
			MacFrameLayout = FLAYOUT_HOST_565;
      VideoMonitor.mode = VMODE_16BIT;
      break;
    case 24:
      MacFrameLayout = FLAYOUT_HOST_888;
      VideoMonitor.mode = VMODE_32BIT;
      break;
    case 32:
      MacFrameLayout = FLAYOUT_HOST_888;
      VideoMonitor.mode = VMODE_32BIT;
      break;
  }
  VideoMonitor.x = width;
  VideoMonitor.y = height;
  VideoMonitor.bytes_per_row = bytes_per_row;

	for( int i=0; i<MAX_SCAN_LINES; i++ ) {
		scanlines[i] = bytes_per_row * i;
	}
}

static int calc_bytes_per_row( int width, int depth )
{
  int result;
  depth = ((depth + 7) / 8) * 8;
  result = width * depth / 8;
  return(result);
}

// Init window mode
static bool init_window(int width, int height, int depth_mac, int depth_win)
{
  bool retval = false;

  HDC hScreenDC = GetDC(hMainWnd);
  if(hScreenDC) {
    hScreenBitmap = CreateBitmap(VideoMonitor.x, VideoMonitor.y, 1, depth_win, NULL);
    hdcBits = CreateCompatibleDC(hScreenDC);
    ReleaseDC(hMainWnd,hScreenDC);
  }

  retval = (hScreenBitmap && hdcBits && the_buffer && lppal);

  if(!retval) {
    if(hScreenBitmap) {
      DeleteObject(hScreenBitmap);
      hScreenBitmap = 0;
    }
    if(hdcBits) {
      DeleteDC(hdcBits);
      hdcBits = 0;
    }
		if(lppal) {
			free(lppal);
			lppal = 0;
		}
  }
	
	MacFrameBaseHost = the_buffer;

  return ( retval );
}

LPDIRECTDRAW lpDD = 0;
DDSURFACEDESC ddsd;
LPDIRECTDRAWSURFACE lpDDSPrimary = 0;
LPDIRECTDRAWPALETTE	lpDDPal = 0;

#define DDPALETTE (DDPCAPS_ALLOW256|DDPCAPS_INITIALIZE|DDPCAPS_8BIT)

static HRESULT restoreAll( void )
{
  HRESULT	ddrval;

  ddrval = lpDDSPrimary->Restore();
	if( ddrval == DD_OK ) {
		if(lpDDPal) lpDDPal->Release();
		lpDD->CreatePalette( DDPALETTE, lppal->palPalEntry, &lpDDPal, NULL );
		if(lpDDPal) lpDDSPrimary->SetPalette(lpDDPal);
	}
  return ddrval;
}

static void finiObjects( void )
{
	if( lpDD ) {
		// (void)lpDD->FlipToGDISurface();
		if(fb_signal) {
			fb_command( FB_QUIT, 0, 0 );
			CloseHandle(fb_signal);
			CloseHandle(fb_reply);
			fb_signal = 0;
			fb_reply = 0;
		}
		if( lpDDSPrimary ) {
			lpDDSPrimary->Release();
			lpDDSPrimary = NULL;
		}
		if( lpDDPal ) {
			lpDDPal->Release();
			lpDDPal = NULL;
		}
		lpDD->Release();
		lpDD = NULL;
	}
}

// Init DX display
static bool init_dx( int width, int height, int depth_mac, int depth_win )
{
	HRESULT ddrval;
	int i;

  ddrval = DirectDrawCreate( NULL, &lpDD, NULL );
  if( ddrval != DD_OK ) {
    finiObjects();
		return(false);
  }
  ddrval = lpDD->SetCooperativeLevel( hMainWnd, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN|DDSCL_ALLOWREBOOT/*|DDSCL_NOWINDOWCHANGES*/ );
  if( ddrval != DD_OK ) {
    finiObjects();
		return(false);
  }
  ddrval = lpDD->SetDisplayMode( width, height, depth_win );
  if(ddrval != DD_OK) {
    finiObjects();
		return(false);
  }
	memset( &ddsd, 0, sizeof(ddsd) );
  ddsd.dwSize = sizeof( ddsd );
  ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
  ddsd.ddsCaps.dwCaps = 
			DDSCAPS_VIDEOMEMORY |
			DDSCAPS_PRIMARYSURFACE |
			DDSCAPS_FLIP |
			DDSCAPS_COMPLEX;
  ddsd.dwBackBufferCount = 1;
  ddrval = lpDD->CreateSurface( &ddsd, &lpDDSPrimary, NULL );
  if( ddrval != DD_OK ) {
    finiObjects();
		return(false);
  }
	if(depth_mac == 8) {
		for ( i = 0; i < 256; i++ ) {
			lppal->palPalEntry[i].peRed = 0;
			lppal->palPalEntry[i].peGreen = 0;
			lppal->palPalEntry[i].peBlue = 0;
			lppal->palPalEntry[i].peFlags = D3DPAL_FREE;
		}
		lpDD->CreatePalette( DDPALETTE, lppal->palPalEntry, &lpDDPal, NULL );
		if(lpDDPal) {
			lpDDSPrimary->SetPalette(lpDDPal);
		}
	}

	// speedup: lookup table
	ddrval = lpDDSPrimary->GetSurfaceDesc( &ddsd );
  if( ddrval != DD_OK ) {
    finiObjects();
		return(false);
  }
	for( i=0; i<MAX_SCAN_LINES; i++ ) {
		dx_scanlines[i] = ddsd.lPitch * i;
	}

	ddrval = lpDDSPrimary->Lock(NULL, &ddsd, 0, NULL);
	if( ddrval == DD_OK ) {
		memset( (uint8 *)ddsd.lpSurface, 0, ddsd.lPitch*VideoMonitor.y );
		lpDDSPrimary->Unlock(0);
	}

  MacFrameBaseHost = the_buffer;

	bPalette = false;

  return true;
}

static bool init_fb( int width, int height, int depth_mac, int depth_win )
{
	HRESULT ddrval;
	int i;

  ddrval = DirectDrawCreate( NULL, &lpDD, NULL );
  if( ddrval != DD_OK ) {
    finiObjects();
		return(false);
  }
  ddrval = lpDD->SetCooperativeLevel( hMainWnd, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN|DDSCL_ALLOWREBOOT );
  if( ddrval != DD_OK ) {
    finiObjects();
		return(false);
  }
  ddrval = lpDD->SetDisplayMode( width, height, depth_win );
  if(ddrval != DD_OK) {
    finiObjects();
		return(false);
  }
	memset( &ddsd, 0, sizeof(ddsd) );
  ddsd.dwSize = sizeof( ddsd );
  ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
  ddsd.ddsCaps.dwCaps = 
			DDSCAPS_VIDEOMEMORY |
			DDSCAPS_PRIMARYSURFACE |
			DDSCAPS_FLIP |
			DDSCAPS_COMPLEX;
  ddsd.dwBackBufferCount = 1;
  ddrval = lpDD->CreateSurface( &ddsd, &lpDDSPrimary, NULL );
  if( ddrval != DD_OK ) {
    finiObjects();
		return(false);
  }
	if(depth_mac == 8) {
		for ( i = 0; i < 256; i++) {
			lppal->palPalEntry[i].peRed = 0;
			lppal->palPalEntry[i].peGreen = 0;
			lppal->palPalEntry[i].peBlue = 0;
			lppal->palPalEntry[i].peFlags = D3DPAL_FREE;
		}
		lpDD->CreatePalette( DDPALETTE, lppal->palPalEntry, &lpDDPal, NULL );
		if(lpDDPal) {
			lpDDSPrimary->SetPalette(lpDDPal);
		}
	}

  MacFrameBaseHost = the_buffer;

	bPalette = false;

  return true;
}

char *get_wnd_class_name(void)
{
	return(window_class);
}

void save_window_position( LPRECT r )
{
	char entry[100], position[100];

	sprintf( entry, "%dx%d", r->right - r->left, r->bottom - r->top );
	sprintf( position, "%d,%d", r->left, r->top );
	WritePrivateProfileString( "Window Positions", entry, position, "BasiliskII.ini" );
}

void save_ontop( void )
{
	WritePrivateProfileString( "Window Positions", "AlwaysOnTop", is_ontop ? "true" : "false", "BasiliskII.ini" );
}

void get_ontop( void )
{
	char tmp[100];

	GetPrivateProfileString( "Window Positions", "AlwaysOnTop", "false", tmp, sizeof(tmp), "BasiliskII.ini" );
	is_ontop = stricmp(tmp,"true") == 0;
}

void get_window_position( int width, int height, int *x, int *y )
{
	char entry[100], position[100];
	int left, top;

	sprintf( entry, "%dx%d", width, height );
	GetPrivateProfileString( "Window Positions", entry, "", position, sizeof(position), "BasiliskII.ini" );
	if(2 == sscanf(position, "%d,%d", &left, &top )) {
		*x = left;
		*x = left;
	} else {
		*x = 0;
		*y = 0;
	}
}

static bool create_window( int width, int height )
{
	WNDCLASSEX wc;
	int x, y, w, h;

	memset( &wc, 0, sizeof(wc) );
	wc.cbSize = sizeof(wc);
	wc.hbrBackground = (HBRUSH)GetStockObject(GRAY_BRUSH);
	wc.hInstance = hInst;
	wc.lpfnWndProc = MainWndProc;
	wc.lpszClassName = window_class;
	wc.style = CS_HREDRAW|CS_VREDRAW|CS_BYTEALIGNCLIENT|CS_BYTEALIGNWINDOW;

  if(is_windowed) {
		w = width;
		h = height;
		get_window_position( width, height, &x, &y );
	} else {
		w = CW_USEDEFAULT;
		h = CW_USEDEFAULT;
		x = y = 0;
	}
  /*if(display_type == DISPLAY_WINDOW)*/ wc.style |= CS_DBLCLKS;

	RegisterClassEx(&wc);

	hMainWnd = CreateWindowEx(
		WS_EX_APPWINDOW,
		window_class,
		GetString(STR_WINDOW_TITLE),
		WS_POPUP,
		x,
		y,
		w,
		h,
		HWND_DESKTOP,
		NULL,
		hInst,
		NULL
	);
	if(!hMainWnd) {
		ErrorAlert("Failed to create main window.");
		return(false);
	} else {
		if(display_type == DISPLAY_WINDOW) {
			get_ontop();
			if(is_ontop) {
				SetWindowPos( hMainWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOSIZE|SWP_NOMOVE );
			}
		}
		return(true);
	}
}

bool VideoInit(void)
{
  int width, height, depth_mac, depth_win;
	
  if(screen_inited) return 0;

	is_windowed = false;

  lppal = (LPLOGPALETTE)malloc( sizeof(LOGPALETTE) + sizeof(PALETTEENTRY) * 256 );
  lppal->palVersion = 0x300;
  lppal->palNumEntries = (WORD)256;

  InitializeCriticalSection( &draw_csection );
  InitializeCriticalSection( &fb_csection );

	width  = GetSystemMetrics(SM_CXSCREEN);
	height = GetSystemMetrics(SM_CYSCREEN);

  HDC hScreenDC = GetDC(GetDesktopWindow());
	if(hScreenDC) {
	  depth_win = GetDeviceCaps(hScreenDC,BITSPIXEL);
		if(depth_win == 16) {
			// The default is the faster one of these.
			// You can override it with screen command.
			depth_mac = 15;
		} else {
			depth_mac = depth_win;
		}
		ReleaseDC(GetDesktopWindow(),hScreenDC);
	}

  display_type = DISPLAY_WINDOW;

	int w_user, h_user, bits_user, got_args;
  const char *mode_str = PrefsFindString("screen");

  if (mode_str && *mode_str) {
    if (strncmp(mode_str, "win",3) == 0) {
      display_type = DISPLAY_WINDOW;
			got_args = sscanf(mode_str, "win/%d/%d/%d", &w_user, &h_user, &bits_user);
    } else if (strncmp(mode_str, "dx",2) == 0) {
      display_type = DISPLAY_DX;
			got_args = sscanf(mode_str, "dx/%d/%d/%d", &w_user, &h_user, &bits_user);
    } else if (strncmp(mode_str, "fb",2) == 0) {
			// Note that this mode cannot be used under Win9x. I'm not
			// disabling it without a good reason. It grabs the Win16Lock
			// for an extended period of time. Try it if you want under Win9x,
			// it will hang your system.
			if(win_os == VER_PLATFORM_WIN32_NT) {
				display_type = DISPLAY_FB;
			}
			got_args = sscanf(mode_str, "fb/%d/%d/%d", &w_user, &h_user, &bits_user);
    }
  }

	if(got_args >= 2) {
		// if either one is zero, the user is trying to define depth only.
		if(w_user || h_user) {
			if(w_user == 0) w_user = width;
			if(h_user == 0) h_user = height;
			if(w_user > 8192) w_user = 8192;
			if(h_user > 8192) h_user = 8192;
			if(display_type == DISPLAY_WINDOW) {
				if(w_user < 80) w_user = 80;
				if(h_user < 80) h_user = 80;
				if(w_user != width || h_user != height) {
					is_windowed = true;
				}
			} else {
				if(w_user < 640) w_user = 640;
				if(h_user < 480) h_user = 480;
			}
			width  = w_user;
			height = h_user;
		}
	}
	if(got_args >= 3 && bits_user != 0 && display_type != DISPLAY_WINDOW) {
		if(bits_user > 32) bits_user = 32;
		if(bits_user < 8) bits_user = 8;
		depth_mac = depth_win = bits_user;
		if(bits_user == 15) {
			depth_mac = 15;
			depth_win = 16;
		}
	}

  if(depth_mac == 24) depth_mac = 32;

	if(mem_8_only) {
		if(depth_mac != 8) {
			ErrorAlert("This is a special version of BasiliskII that runs only in 8 bit color mode.");
			QuitEmulator();
		}
		if(display_type == DISPLAY_FB) {
			// Linear frame buffer address is not predictable for my purposes.
			// Use normal d3d as a fallback
			display_type = DISPLAY_DX;
		}
	}

  frame_skip = PrefsFindInt32("frameskip");
  if (frame_skip <= 0) frame_skip = 1;

  pbmi = (LPBITMAPINFO)malloc( sizeof(BITMAPINFO) + 256 * sizeof(RGBQUAD) );
  if(!pbmi) QuitEmulator();

  memset( pbmi, 0, sizeof(BITMAPINFO) + 256 * sizeof(RGBQUAD) );

  pbmi = (LPBITMAPINFO)malloc( sizeof(BITMAPINFO) + 256 * sizeof(RGBQUAD) );
  pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  pbmi->bmiHeader.biWidth = width;
  pbmi->bmiHeader.biPlanes = 1;
  pbmi->bmiHeader.biBitCount = depth_mac==15 ? 16 : depth_mac;
  pbmi->bmiHeader.biCompression = 0;
  pbmi->bmiHeader.biXPelsPerMeter = 0;
  pbmi->bmiHeader.biYPelsPerMeter = 0;
  pbmi->bmiHeader.biClrUsed = depth_mac == 8 ? 256 : 0;
  pbmi->bmiHeader.biClrImportant = depth_mac == 8 ? 256 : 0;

	for( int i=0; i<256; i++ ) {
    pbmi->bmiColors[i].rgbReserved = 0; //PC_NOCOLLAPSE|PC_RESERVED;
	}

  ADBSetRelMouseMode(false);
  set_video_monitor( width, height, depth_win, depth_mac, calc_bytes_per_row(width,depth_mac) );

	int sz = width * calc_bytes_per_row(height,depth_mac) + VM_PAGE_BYTES*16;

	if(mem_8_only) {
		if(!the_buffer) {
			ErrorAlert("Internal error: the_buffer is not mapped to process address space.");
			QuitEmulator();
		}
	} else {
		the_buffer = (uint8 *)VirtualAlloc( 0, sz, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
	}
	if(!the_buffer) return false;
	memset( the_buffer, 0, sz );

  VideoMonitor.mac_frame_base = MacFrameBaseMac;

	if(!create_window(width,height)) QuitEmulator();
	
	if(is_windowed) {
		ShowWindow(hMainWnd, SW_SHOWNORMAL);
	} else {
		ShowWindow(hMainWnd, SW_MAXIMIZE);
	}

  // Initialize according to display type
	// If DirectX init fails use WINAPI as a fallback.
  switch (display_type) {
    case DISPLAY_WINDOW:
      if (!init_window(width, height, depth_mac, depth_win))
        return false;
      break;
    case DISPLAY_FB:
      if (!init_fb(width, height, depth_mac, depth_win)) {
				display_type = DISPLAY_WINDOW;
				DestroyWindow(hMainWnd);
				final_desktop();
				if(!create_window(width,height)) QuitEmulator();
				ShowWindow(hMainWnd, SW_MAXIMIZE);
	      if (!init_window(width, height, depth_mac, depth_win))
					return false;
		    WarningAlert("Frame buffer initialization failed. Using Windows GDI API.");
			}
      break;
    case DISPLAY_DX:
      if (!init_dx(width, height, depth_mac, depth_win)) {
				display_type = DISPLAY_WINDOW;
	      if (!init_window(width, height, depth_mac, depth_win))
					return false;
		    WarningAlert("DirectX initialization failed. Using Windows GDI API.");
			}
      break;
  }

	ShowCursor(FALSE);

	nt_dx5_palette_hack = 
			PrefsFindBool("ntdx5hack") && 
			(win_os == VER_PLATFORM_WIN32_NT) && 
			(display_type != DISPLAY_WINDOW);

	// kludge
  if(display_type != DISPLAY_WINDOW) {
		ShowWindow(hMainWnd, SW_MINIMIZE);
		ShowWindow(hMainWnd, SW_MAXIMIZE);
		SetFocus(hMainWnd);
		SetActiveWindow(hMainWnd);
	}

#ifdef USE_FAULTS
	// Direct frame buffer access does not need the fault handler
  if(display_type != DISPLAY_FB) {
		Screen_Setup_fault_handler();
	}
#endif //USE_FAULTS

  // Set variables for UAE memory mapping
  MacFrameSize = VideoMonitor.bytes_per_row * VideoMonitor.y;
	// This is set in init_*()
  // MacFrameBaseHost = the_buffer;

  screen_inited = 1;

	unsigned int tid;
  if(display_type == DISPLAY_WINDOW) {
		_beginthreadex( 0, 0, redraw_thread, 0, 0, &tid );
  } else if(display_type == DISPLAY_FB) {
		fb_signal = CreateSemaphore( 0, 0, 1, NULL);
		fb_reply = CreateSemaphore( 0, 0, 1, NULL);
		_beginthreadex( 0, 0, redraw_thread_fb, 0, 0, &tid );
  } else if(display_type == DISPLAY_DX) {
		_beginthreadex( 0, 0, redraw_thread_d3d, 0, 0, &tid );
	}

  return true;
}

bool is_screen_inited(void)
{
	return(screen_inited);
}


/*
 *  Deinitialization
 */

void VideoExit(void)
{
	grab_draw_mutex();

  screen_inited = 0;
	dx_blits_allowed = 0;

	finiObjects();

  if(the_buffer && !mem_8_only) {
    VirtualFree( the_buffer, 0, MEM_RELEASE  );
		the_buffer = 0;
  }
  if(lppal) {
		free(lppal);
		lppal = 0;
	}
  if(hPalMain) {
    DeleteObject(hPalMain);
    hPalMain = 0;
  }
  if(hPalBits) {
    DeleteObject(hPalBits);
    hPalBits = 0;
  }
  if(pbmi) {
    free(pbmi);
    pbmi = 0;
  }
  if(hScreenBitmap) {
    DeleteObject(hScreenBitmap);
    hScreenBitmap = 0;
  }
  if(hdcBits) {
    DeleteDC(hdcBits);
    hdcBits = 0;
  }
	release_draw_mutex();

	final_desktop();

	// Let the system destroy this one
  // DeleteCriticalSection( &draw_csection );

  // DeleteCriticalSection( &fb_csection );
}

#pragma pack(1)
typedef struct {
	BYTE red, green, blue;
} macpal_type;
#pragma pack()

typedef struct {
	int32 diff;
	uint32 inx;
} macpal_inx;

int __cdecl palsort( const void *a, const void *b)
{
	return( ((macpal_inx *)b)->diff - ((macpal_inx *)a)->diff );
}

// not used, for future reference
void setup_palette_entries( HDC hdc, macpal_type pal[256])
{
	int i;
	macpal_inx diff[256];
	BYTE red, green, blue;

	diff[0].inx = 0;
	diff[0].diff = 50000;
	diff[255].inx = 255;
	diff[255].diff = 50000;

	GetSystemPaletteEntries( hdc, 0, 256, lppal->palPalEntry );

  for( i=1; i<255; i++ ) {
		diff[i].inx = i;
		red = pal[i].red;
		green = pal[i].green;
		blue = pal[i].blue;
		if( 0 && (red == 0 || green == 0 || blue == 0 || 
		   red == 255 || green == 255 || blue == 255) )
		{
			diff[i].diff = 50000;
		} else {
			diff[i].diff = 
				abs( (int)pal[i-1].red - (int)pal[i].red) + 
				abs( (int)pal[i-1].green - (int)pal[i].green) + 
				abs( (int)pal[i-1].blue - (int)pal[i].blue);
		}
	}
	qsort( (void *)diff, 256, sizeof(diff[0]), palsort );

  for (i=10; i<246; i++) {
    lppal->palPalEntry[i+10].peRed = pal[diff[i].inx].red;
    lppal->palPalEntry[i+10].peGreen = pal[diff[i].inx].green;
    lppal->palPalEntry[i+10].peBlue = pal[diff[i].inx].blue;
    lppal->palPalEntry[i+10].peFlags = PC_NOCOLLAPSE;
    pbmi->bmiColors[i].rgbReserved = 0; // PC_RESERVED|PC_NOCOLLAPSE
  }
}

void setup_palette_entries(uint8 *pal)
{
	uint16 i, j;
	uint16 *iptr = (uint16 *)&pbmi->bmiColors[0];

  for (i=0,j=0; i<256; i++,j+=3) {
    lppal->palPalEntry[i].peRed = pal[j];
    lppal->palPalEntry[i].peGreen = pal[j+1];
    lppal->palPalEntry[i].peBlue = pal[j+2];
    lppal->palPalEntry[i].peFlags = PC_NOCOLLAPSE; // PC_RESERVED|PC_NOCOLLAPSE
		if(bPalette) {
			*iptr++ = i;
		} else {
			pbmi->bmiColors[i].rgbRed = pal[j];
			pbmi->bmiColors[i].rgbGreen = pal[j+1];
			pbmi->bmiColors[i].rgbBlue = pal[j+2];
			pbmi->bmiColors[i].rgbReserved = 0;
		}
  }
}

void setup_palette_entries_dd3(uint8 *pal)
{
	uint16 i, j;

  for (i=0,j=0; i<256; i++,j+=3) {
    lppal->palPalEntry[i].peRed = pal[j];
    lppal->palPalEntry[i].peGreen = pal[j+1];
    lppal->palPalEntry[i].peBlue = pal[j+2];
  }
}

// Need to be inside critical section already
static void __inline__ set_very_dirty( void )
{
#ifdef USE_FAULTS
	mainBuffer.dirty = 1;
	mainBuffer.top = 0;
	mainBuffer.bottom = SCREEN_MAX_Y;
	mainBuffer.unprotect_start = mainBuffer.memStart;
	mainBuffer.unprotect_bytes = mainBuffer.memLength;
#endif
}

void video_set_palette2(uint8 *pal)
{
	HDC hScreenDC;
	HRESULT ddrval;
	int dt = display_type;

	if(!screen_inited) return;

	grab_draw_mutex();

	if(nt_dx5_palette_hack && dt == DISPLAY_DX) dt = DISPLAY_WINDOW;

  if(screen_bit_depth == 8) {
		switch(dt) {
			case DISPLAY_DX:
				if(lpDDPal) {
					setup_palette_entries_dd3(pal);
					for(;;) {
						ddrval = lpDDPal->SetEntries( 0, 0, 256, lppal->palPalEntry );
						if( ddrval == DD_OK ) break;
						if( ddrval == DDERR_SURFACELOST ) {
							ddrval = restoreAll();
						}
						if( ddrval != DD_OK && ddrval != DD_OK && ddrval != DDERR_WASSTILLDRAWING && ddrval != DDERR_SURFACEBUSY ) break;
					}
				}
				set_very_dirty();
				break;
			case DISPLAY_FB:
				if(lpDDPal) {
					fb_command( FB_SETPALETTE, pal, 768 );
				}
				break;
			default:
				hScreenDC = GetDC(hMainWnd);
				if(hScreenDC) {
					HPALETTE hold;

					if(hPalMain) DeleteObject(hPalMain);
					if(hPalBits) DeleteObject(hPalBits);

					setup_palette_entries( pal );
					hPalMain = CreatePalette(lppal);
					hold = SelectPalette( hScreenDC, hPalMain, FALSE );
					RealizePalette(hScreenDC);
					SelectPalette( hScreenDC, hold, FALSE );

					hPalBits = CreatePalette(lppal);
					hold = SelectPalette( hdcBits, hPalBits, FALSE );
					RealizePalette(hdcBits);
					SelectPalette( hdcBits, hold, FALSE );

					ReleaseDC(hMainWnd,hScreenDC);
					set_very_dirty();
				}
				break;
		}
  }
  
	release_draw_mutex();
}

/*
 *  Set palette
 */

void video_set_palette(uint8 *pal)
{
	memcpy( current_mac_palette, pal, sizeof(current_mac_palette) );
	video_set_palette2(pal);
}

int update_palette()
{
	int i = 0;
	HRESULT ddrval;
	int dt = display_type;

	if(!screen_inited) return(0);

	grab_draw_mutex();

	if(nt_dx5_palette_hack && dt == DISPLAY_DX) dt = DISPLAY_WINDOW;

  switch(dt) {
		case DISPLAY_DX:
			if(lpDDPal) {
				for(;;) {
					ddrval = lpDDSPrimary->SetPalette(lpDDPal);
					if( ddrval == DD_OK ) break;
					if( ddrval == DDERR_SURFACELOST ) {
						ddrval = restoreAll();
					}
					if( ddrval != DD_OK && ddrval != DDERR_WASSTILLDRAWING && ddrval != DDERR_SURFACEBUSY ) break;
				}
			}
			i = 1;
			set_very_dirty();
			break;
		case DISPLAY_FB:
			if(lpDDPal) {
				fb_command( FB_UPDATEPALETTE, 0, 0 );
			}
			i = 1;
			break;
		default:
			if(hPalMain && hPalBits) {
				HDC hScreenDC = GetDC(hMainWnd);
				if(hScreenDC) {
					HPALETTE hold;
					// UnrealizeObject(hPalMain);
					// UnrealizeObject(hPalBits);
					hold = SelectPalette( hScreenDC, hPalMain, FALSE );
					i = RealizePalette(hScreenDC);
					SelectPalette( hScreenDC, hold, FALSE );
					hold = SelectPalette( hdcBits, hPalBits, FALSE );
					i += RealizePalette(hdcBits);
					SelectPalette( hdcBits, hold, FALSE );
					if(i) {
						set_very_dirty();
					}
					ReleaseDC(hMainWnd,hScreenDC);
				}
			}
			break;
	}
	release_draw_mutex();
	return( i );
}

#ifdef USE_FAULTS
void Screen_Draw(HDC hScreenDC)
{
  BYTE *pbyScreen;
  DWORD OldProtect;
  HBITMAP oldBitmap;
  int y, line_count;
	HPALETTE hold1 = 0, hold2 = 0;

  if(the_buffer && screen_inited && mainBuffer.dirty) {
    mainBuffer.dirty = 0;

    VirtualProtect( (LPVOID)mainBuffer.unprotect_start, mainBuffer.unprotect_bytes, PAGE_READONLY, &OldProtect );

    y = mainBuffer.top;
		line_count = mainBuffer.bottom - y;
    pbyScreen = &((BYTE *)mainBuffer.memBase)[scanlines[y]];

	  if(display_type == DISPLAY_DX) {
			if(dx_blits_allowed) {
				ddsd.dwSize = sizeof(ddsd);
				for(;;) {
					HRESULT	ddrval = lpDDSPrimary->Lock(NULL, &ddsd, 0, NULL);
					if( ddrval == DD_OK ) {
				    uint8 *p2 = &((BYTE *)ddsd.lpSurface)[dx_scanlines[y]];
						int n1 = VideoMonitor.bytes_per_row;
						int n2 = ddsd.lPitch;

						if(n2 == 0) n2 = n1;
						if(n1 == n2) {
							memcpy( p2, pbyScreen, scanlines[line_count] );
						} else {
							for( int i=line_count; i; i-- ) {
								memcpy( p2, pbyScreen, n2 );
								p2 += n2;
								pbyScreen += n1;
							}
						}
						lpDDSPrimary->Unlock(0);
						break;
					}
					if( ddrval == DDERR_SURFACELOST ) {
						ddrval = restoreAll();
					}
					if( ddrval != DD_OK && ddrval != DDERR_WASSTILLDRAWING && ddrval != DDERR_SURFACEBUSY ) break;
				}
			}
			return;
		}

		switch(screen_bit_depth) {
			case 8:
				if(hPalMain) {
					hold1 = SelectPalette( hScreenDC, hPalMain, FALSE );
					hold2 = SelectPalette( hdcBits, hPalBits, FALSE );
				}
				pbmi->bmiHeader.biHeight = -line_count;
				pbmi->bmiHeader.biSizeImage = scanlines[line_count];
				SetDIBits( hdcBits, hScreenBitmap, 0, line_count, pbyScreen, pbmi, 
					bPalette ? DIB_PAL_COLORS : DIB_RGB_COLORS );
				break;
			case 15:
			case 16:
				// SetBitmapBits(hScreenBitmap, scanlines[line_count], pbyScreen );
				pbmi->bmiHeader.biHeight = -line_count;
				pbmi->bmiHeader.biSizeImage = scanlines[line_count];
				SetDIBits( hdcBits, hScreenBitmap, 0, line_count, pbyScreen, pbmi, DIB_RGB_COLORS );
				break;
			case 24:
				pbmi->bmiHeader.biHeight = -line_count;
				pbmi->bmiHeader.biSizeImage = (line_count*VideoMonitor.x) << 2;
				SetDIBits( hdcBits, hScreenBitmap, 0, line_count, pbyScreen, pbmi, DIB_RGB_COLORS );
				break;
			case 32:
				pbmi->bmiHeader.biHeight = -line_count;
				pbmi->bmiHeader.biSizeImage = scanlines[line_count];
				SetDIBits( hdcBits, hScreenBitmap, 0, line_count, pbyScreen, pbmi, DIB_RGB_COLORS );
				break;
		}

    oldBitmap = SelectObject(hdcBits,hScreenBitmap);
    BitBlt(hScreenDC, 0, y, VideoMonitor.x, line_count, hdcBits, 0, 0, SRCCOPY );
		if(hPalMain) {
			SelectPalette( hScreenDC, hold1, FALSE );
			SelectPalette( hdcBits, hold2, FALSE );
		}
    SelectObject(hdcBits,oldBitmap);
  }
}

void Screen_Draw_All(HWND hWnd)
{
  HDC hdc;
  PAINTSTRUCT ps;

	// Even if you remove the Screen_Draw() below, you need to keep
	// this mutex. It protects also the state variables.
	grab_draw_mutex();

	// This is not stricly necessary since the redraw thread would paint
	// the screen anyway. It just makes a more pleasant effect when the
	// Mac windows is updated immediately.
  switch(display_type) {
		case DISPLAY_WINDOW:
		  hdc = BeginPaint(hWnd, &ps);
			set_very_dirty();
			//Screen_Draw(hdc);
		  EndPaint(hWnd, &ps);
			break;
		case DISPLAY_DX:
		  hdc = BeginPaint(hWnd, &ps);
		  EndPaint(hWnd, &ps);
			if(dx_blits_allowed) {
				set_very_dirty();
				//Screen_Draw(0);
			}
			break;
		case DISPLAY_FB:
		  hdc = BeginPaint(hWnd, &ps);
		  EndPaint(hWnd, &ps);
			break;
	}

	release_draw_mutex();
}
#else
void Screen_Draw(HDC hScreenDC)
{
	HPALETTE hold1 = 0, hold2 = 0;

  if(!screen_inited) return;

	// This code is for debugging only and not necessarily up to date.
	// See the release code above 

	if(display_type == DISPLAY_DX) {
		if(dx_blits_allowed) {
			ddsd.dwSize = sizeof(ddsd);
			for(;;) {
				HRESULT	ddrval = lpDDSPrimary->Lock(NULL, &ddsd, 0, NULL);
				if( ddrval == DD_OK ) {
					uint8 *p2 = (uint8 *)ddsd.lpSurface;
					uint8 *p1 = (uint8 *)the_buffer;
					int n1 = VideoMonitor.bytes_per_row;
					int n2 = ddsd.lPitch;
					if(n2 == 0) n2 = n1;
					for( int i=VideoMonitor.y; i; i-- ) {
						memcpy( p2, p1, n2 );
						p2 += n2;
						p1 += n1;
					}
					lpDDSPrimary->Unlock(0);
					break;
				}
				if( ddrval == DDERR_SURFACELOST ) {
					ddrval = restoreAll();
					// if( ddrval != DD_OK ) break;
				}
				if( ddrval != DD_OK && ddrval != DDERR_WASSTILLDRAWING && ddrval != DDERR_SURFACEBUSY ) break;
			}
		}
		return;
	}

	switch(screen_bit_depth) {
		case 8:
			if(hPalMain) {
				hold1 = SelectPalette( hScreenDC, hPalMain, FALSE );
				hold2 = SelectPalette( hdcBits, hPalBits, FALSE );
			}
			pbmi->bmiHeader.biHeight = -(int)VideoMonitor.y;
			pbmi->bmiHeader.biSizeImage = scanlines[VideoMonitor.y];
			SetDIBits( hdcBits, hScreenBitmap, 0, VideoMonitor.y, the_buffer, pbmi, 
					bPalette ? DIB_PAL_COLORS : DIB_RGB_COLORS );
			break;
		case 24:
			pbmi->bmiHeader.biHeight = -(int)VideoMonitor.y;
			pbmi->bmiHeader.biSizeImage = VideoMonitor.y*VideoMonitor.x*4;
			SetDIBits( hdcBits, hScreenBitmap, 0, VideoMonitor.y, the_buffer, pbmi, DIB_RGB_COLORS );
			break;
		default:
	    SetBitmapBits(hScreenBitmap, scanlines[VideoMonitor.y], the_buffer );
	}

  HBITMAP oldBitmap = SelectObject(hdcBits,hScreenBitmap);
  BitBlt(hScreenDC, 0, 0, VideoMonitor.x, VideoMonitor.y, hdcBits, 0, 0, SRCCOPY );
	if(hPalMain) {
		SelectPalette( hScreenDC, hold1, FALSE );
		SelectPalette( hdcBits, hold2, FALSE );
	}
  SelectObject(hdcBits,oldBitmap);
}

void Screen_Draw_All(HWND hWnd)
{
  HDC hdc;
  PAINTSTRUCT ps;

	grab_draw_mutex();

  switch(display_type) {
		case DISPLAY_WINDOW:
		  hdc = BeginPaint(hWnd, &ps);
			Screen_Draw(hdc);
		  EndPaint(hWnd, &ps);
			break;
		case DISPLAY_DX:
		  hdc = BeginPaint(hWnd, &ps);
		  EndPaint(hWnd, &ps);
			if(dx_blits_allowed) {
				Screen_Draw(0);
			}
			break;
		case DISPLAY_FB:
		  hdc = BeginPaint(hWnd, &ps);
		  EndPaint(hWnd, &ps);
			break;
	}

	release_draw_mutex();
}
#endif //USE_FAULTS

/*
 *  Thread for screen refresh
 */

unsigned int WINAPI redraw_thread(LPVOID param)
{
	set_desktop();

#ifndef USE_FAULTS
  int tick_counter = 0;
#endif //USE_FAULTS
#ifdef SHOW_REAL_FRAME_RATE
  DWORD showtime = GetTickCount() + 1000;
  int frame_counter = 0;
#endif //SHOW_REAL_FRAME_RATE
  HDC hScreenDC = 0;

  while(screen_inited) {
		Sleep(SLEEP_TICKS);
#ifndef USE_FAULTS
    if (++tick_counter >= frame_skip) {
      tick_counter = 0;
#endif
			grab_draw_mutex();
			hScreenDC = GetDC(hMainWnd);
			if(hScreenDC) {
				Screen_Draw(hScreenDC);
				ReleaseDC(hMainWnd,hScreenDC);
			}
			release_draw_mutex();
#ifndef USE_FAULTS
    }
#endif
#ifdef SHOW_REAL_FRAME_RATE
		frame_counter++;
		if(GetTickCount() >= showtime) {
			grab_draw_mutex();
      hScreenDC = GetDC(hMainWnd);
      if(hScreenDC) {
				char b[100];
				wsprintf( b, "Framerate: %d  ", frame_counter );
				TextOut( hScreenDC, 400, 0, b, strlen(b) );
        ReleaseDC(hMainWnd,hScreenDC);
      }
			release_draw_mutex();
			frame_counter = 0;
			showtime = GetTickCount() + 1000;
		}
#endif //SHOW_REAL_FRAME_RATE
  }
	_endthreadex( 0 );
  return 0;
}

unsigned int WINAPI redraw_thread_d3d(LPVOID param)
{
	set_desktop();

#ifndef USE_FAULTS
  int tick_counter = 0;
#endif //USE_FAULTS
#ifdef SHOW_REAL_FRAME_RATE
  DWORD showtime = GetTickCount() + 1000;
  int frame_counter = 0;
  HDC hScreenDC = 0;
#endif //SHOW_REAL_FRAME_RATE

  while(screen_inited) {
		Sleep(SLEEP_TICKS);
#ifndef USE_FAULTS
    if (++tick_counter >= frame_skip) {
      tick_counter = 0;
#endif
			grab_draw_mutex();
			Screen_Draw(0);
			release_draw_mutex();
#ifndef USE_FAULTS
    }
#endif
#ifdef SHOW_REAL_FRAME_RATE
		frame_counter++;
		if(GetTickCount() >= showtime) {
			grab_draw_mutex();
      hScreenDC = GetDC(hMainWnd);
      if(hScreenDC) {
				char b[100];
				wsprintf( b, "Framerate: %d  ", frame_counter );
				TextOut( hScreenDC, 400, 0, b, strlen(b) );
        ReleaseDC(hMainWnd,hScreenDC);
      }
			release_draw_mutex();
			frame_counter = 0;
			showtime = GetTickCount() + 1000;
		}
#endif //SHOW_REAL_FRAME_RATE
  }
	_endthreadex( 0 );
  return 0;
}

void VideoInterrupt(void)
{
}

// When using linear frame buffer, graphics data is dumped
// directly on the screen. However, when the user task-switches
// away, we take a snapshot of the screen to "the_buffer" and
// set up background processing screen data to go temporarily 
// to "the_buffer". When the user switches back, we blit the_buffer
// to screen and and resume linear frame buffer operation.
static void fb_blit( bool memorize )
{
	uint8 *p2 = (uint8 *)ddsd.lpSurface;
	uint8 *p1 = (uint8 *)the_buffer;
	int n1 = VideoMonitor.bytes_per_row;
	int n2 = ddsd.lPitch;
	if(n2 == 0) n2 = n1;
	for( int i=VideoMonitor.y; i; i-- ) {
		if(memorize) {
			memcpy( p1, p2, n2 );
		} else {
			memcpy( p2, p1, n2 );
		}
		p2 += n2;
		p1 += n1;
	}
}

// Buffer to pass the new palette to the frame buffer thread
static uint8 fb_buffer[768];


// Linear frame buffer thread
unsigned int WINAPI redraw_thread_fb(LPVOID param)
{
	// Is the frame locked *now*
	bool is_frame_locked = false;

	// Quit command breaks out from looping
	bool looping = true;

	// Indicates whether the next command should lock or not
	bool lock_next = true;

	// Flag used to inhibit locking when we are away
	bool waiting_to_be_locked = false;

  HRESULT	ddrval = 0;


	set_desktop();

	D(bug("fb start\n"));

	// kludge
	ShowWindow(hMainWnd, SW_MINIMIZE);
	ShowWindow(hMainWnd, SW_MAXIMIZE);
	SetFocus(hMainWnd);
	SetActiveWindow(hMainWnd);

	while(looping) {
		if(lock_next) {
			D(bug("locking...\n"));

			for(;;) {
				ddsd.dwSize = sizeof( ddsd );
				ddrval = lpDDSPrimary->Lock(NULL, &ddsd, 0, NULL);
				if( ddrval == DD_OK ) {
					is_frame_locked = true;
					MacFrameBaseHost = (BYTE *)ddsd.lpSurface;
					memory_init();
					if( fb_current_command == FB_LOCK) {
						D(bug("restoring frame...\n"));
						fb_blit(false);
					}
					break;
				}
				if( ddrval == DDERR_SURFACELOST ) {
					ddrval = restoreAll();
				}
				if( ddrval != DD_OK && ddrval != DDERR_WASSTILLDRAWING && ddrval != DDERR_SURFACEBUSY ) {
					D(bug("fb lock failed\n"));
					break;
				}
			}
		}

		if(!waiting_to_be_locked) lock_next = true;

		D(bug("fb waiting...\n"));
		WaitForSingleObject(fb_signal,INFINITE);

		if(is_frame_locked) {
			if( fb_current_command == FB_UNLOCK) {
				D(bug("memorizing frame...\n"));
				fb_blit(true);
			}

			D(bug("unlocking...\n"));
			MacFrameBaseHost = the_buffer;
			memory_init();
			for(;;) {
				ddrval = lpDDSPrimary->Unlock(0);
				if( ddrval == DD_OK ) {
					is_frame_locked = false;
					break;
				}
				if( ddrval != DD_OK && ddrval != DDERR_WASSTILLDRAWING && ddrval != DDERR_SURFACEBUSY ) break;
			}
		}

		if(waiting_to_be_locked) {
			if(fb_current_command != FB_LOCK && fb_current_command != FB_QUIT) {
				fb_current_command = FB_NONE;
				lock_next = false;
			}
		}

		switch( fb_current_command ) {
			case FB_UNLOCK:
				D(bug("FB_UNLOCK\n"));
				lock_next = false;
				waiting_to_be_locked = true;
				break;
			case FB_LOCK:
				D(bug("FB_LOCK\n"));
				waiting_to_be_locked = false;
				lock_next = true;
				break;
			case FB_QUIT:
				D(bug("FB_QUIT\n"));
				looping = false;
				break;
			case FB_SETPALETTE:
				D(bug("FB_SETPALETTE\n"));
				if(is_frame_locked) break;
				if(screen_bit_depth != 8) break;

				// This damn hack again.
				if(nt_dx5_palette_hack) {
					HDC hScreenDC = GetDC(hMainWnd);
					if(hScreenDC) {
						HPALETTE hold;

						if(hPalMain) DeleteObject(hPalMain);
						if(hPalBits) DeleteObject(hPalBits);

						setup_palette_entries( fb_buffer );
						hPalMain = CreatePalette(lppal);
						hold = SelectPalette( hScreenDC, hPalMain, FALSE );
						RealizePalette(hScreenDC);
						SelectPalette( hScreenDC, hold, FALSE );

						hPalBits = CreatePalette(lppal);
						hold = SelectPalette( hdcBits, hPalBits, FALSE );
						RealizePalette(hdcBits);
						SelectPalette( hdcBits, hold, FALSE );

						ReleaseDC(hMainWnd,hScreenDC);
					}
				} else {
					if(lpDDPal) {
						setup_palette_entries_dd3(fb_buffer);
						for(;;) {
							ddrval = lpDDPal->SetEntries( 0, 0, 256, lppal->palPalEntry );
							if( ddrval == DD_OK ) break;
							if( ddrval == DDERR_SURFACELOST ) {
								ddrval = restoreAll();
							}
							if( ddrval != DD_OK && ddrval != DDERR_WASSTILLDRAWING && ddrval != DDERR_SURFACEBUSY ) break;
						}
					}
				}
				break;
			case FB_UPDATEPALETTE:
				D(bug("FB_UPDATEPALETTE\n"));
				if(lpDDPal && !is_frame_locked && screen_bit_depth == 8) {
					for(;;) {
						ddrval = lpDDSPrimary->SetPalette(lpDDPal);
						if( ddrval == DD_OK ) break;
						if( ddrval == DDERR_SURFACELOST ) {
							ddrval = restoreAll();
						}
						if( ddrval != DD_OK && ddrval != DDERR_WASSTILLDRAWING && ddrval != DDERR_SURFACEBUSY ) break;
					}
				}
				break;
		}

		ReleaseSemaphore(fb_reply,1,NULL);
	}

	D(bug("fb thread stop\n"));
	_endthreadex( 0 );

	return(0);
}

void fb_command( uint8 cmd, LPBYTE param, uint32 sz )
{
	D(bug("fb command %ld\n",(long)cmd));

	EnterCriticalSection( &fb_csection );
	fb_current_command = cmd;
	if(param && sz) {
		memcpy( fb_buffer, param, sz );
	}
	ReleaseSemaphore(fb_signal,1,NULL);
	WaitForSingleObject(fb_reply,INFINITE);
	LeaveCriticalSection( &fb_csection );
}


void video_activate( bool activate )
{
  if(display_type == DISPLAY_FB) {
		// fb_command( activate ? FB_LOCK : FB_UNLOCK, 0, 0 );
	} else if(display_type == DISPLAY_DX) {
		dx_blits_allowed = activate;
		if(activate) {
			video_set_palette2(current_mac_palette);
			grab_draw_mutex();
			set_very_dirty();
			Screen_Draw(0);
			release_draw_mutex();
		}
	}
}
